#include "Python.h"
#include "Numeric/arrayobject.h"

static char doc_arrayfrombuffer[] = 
"arrayfrombuffer(buf, typecode='l') makes an array whose data is in buf\n"
"\n"
"This function makes a new Numeric array whose data is stored in\n"
"buf, a read-write buffer object; for example, a read-write mmap\n"
"object.  Any changes to the array will change the data in buf,\n"
"and any changes to the data in buf will change the array.\n"
"\n"
"The typecode is optional; it defaults to 'l', like fromstring's.\n"
"\n"
"This is useful, for example, for storing an array in a\n"
"memory-mapped file or a shared-memory segment, or just for\n"
"avoiding memory-to-memory copying.\n"
"\n"
"I considered supporting read-only buffers, but Numeric doesn't\n"
"support read-only arrays at the moment, and I don't need them for\n"
"my application.\n"
; 
static PyObject *
arrayfrombuffer(PyObject *self, PyObject *args)
{
  PyArrayObject *arrayobj;
  PyObject *bufobj;
  char *type = "l";  /* long int is default */
  void *data;
  int datalen;
  int nitems;
  PyArray_Descr *descr;

  if (!PyArg_ParseTuple(args, "O|s:arrayfrombuffer", &bufobj, &type)) {
    return NULL;
  }
  if (-1 == PyObject_AsWriteBuffer(bufobj, &data, &datalen)) {
    return NULL;
  }
  if (!(descr = PyArray_DescrFromType(*type))) {
    return NULL;
  }
  if (datalen % descr->elsize != 0) {
    PyErr_SetString(PyExc_ValueError, "buffer length must be a multiple of element size");
    return NULL;
  }
  nitems = datalen / descr->elsize;
  /* The Numeric docs say not to do this unless 'data' is eternal,
   * but they're wrong.  See longer comment below. */
  if (!(arrayobj = (PyArrayObject*)PyArray_FromDimsAndData(1, 
							   &nitems, 
							   *type, 
							   data))) {
    return NULL;
  }
  /* This prevents the bufobj from being inadvertently
   * deallocated before the array is, and arranges for the bufobj
   * to be deallocated when the array is if it's not otherwise
   * referenced.  Of course, you can still cause a core dump or
   * incorrect results by invalidating the bufobj's memory before
   * deleting the array --- for example, by calling close() on an
   * mmap object.  Them's the breaks.*/
  arrayobj->base = bufobj;
  Py_INCREF(bufobj);
  /* do I need to incref arrayobj?! fromstring does... */
  return (PyObject*)arrayobj;
}

static PyMethodDef arrayfrombufferfuncs[] = {
  {"arrayfrombuffer", arrayfrombuffer, METH_VARARGS, doc_arrayfrombuffer},
  {NULL, NULL}
};

void
initarrayfrombuffer(void)
{
  (void)Py_InitModule("arrayfrombuffer", arrayfrombufferfuncs);
  import_array();
}
